/*
 *  $Id: alltoallv.c,v 1.16 1994/07/13 15:31:45 lusk Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      See COPYRIGHT in top-level directory.
 */

#ifndef lint
static char vcid[] = "$Id: alltoallv.c,v 1.16 1994/07/13 15:31:45 lusk Exp $";
#endif /* lint */

#include "mpiimpl.h"
#include "coll.h"

/*@

MPI_Alltoallv - Sends data from all to all processes, with a displacement

Input Parameters:
. sendbuf - starting address of send buffer (choice) 
. sendcounts - integer array equal to the group size 
specifying the number of elements to send to each processor 
. sdispls - integer array (of length group size). Entry 
 j  specifies the displacement (relative to sendbuf  from
which to take the outgoing data destined for process  j  
. sendtype - data type of send buffer elements (handle) 
. recvcounts - integer array equal to the group size 
specifying the maximum number of elements that can be received from
each processor 
. rdispls - integer array (of length group size). Entry 
 i  specifies the displacement (relative to recvbuf  at
which to place the incoming data from process  i  
. recvtype - data type of receive buffer elements (handle) 
. comm - communicator (handle) 

Output Parameter:
. recvbuf - address of receive buffer (choice) 
@*/
int MPI_Alltoallv ( sendbuf, sendcnts, sdispls, sendtype, 
                    recvbuf, recvcnts, rdispls, recvtype, comm )
void             *sendbuf;
int              *sendcnts;
int              *sdispls;
MPI_Datatype      sendtype;
void             *recvbuf;
int              *recvcnts;
int              *rdispls; 
MPI_Datatype      recvtype;
MPI_Comm          comm;
{
  int        size, rank, i;
  MPI_Aint   extent;
  int        errno = MPI_SUCCESS;
  MPI_Status status;
  int        flag;
  
  /* Check for invalid arguments */
  if ( MPIR_TEST_COMM(comm,comm) || MPIR_TEST_DATATYPE(comm,sendtype) ||
       MPIR_TEST_DATATYPE(comm,recvtype))
	return MPIR_ERROR(comm, errno, "Error in MPI_ALLTOALLV" ); 

  /* Check for intra-communicator */
  MPI_Comm_test_inter ( comm, &flag );
  if (flag) 
    return MPIR_ERROR(comm, MPI_ERR_COMM,
			  "Inter-communicator invalid in MPI_ALLTOALLV");
  
  /* Get size and rank and switch to collective communicator */
  MPI_Comm_size ( comm, &size );
  MPI_Comm_rank ( comm, &rank );
  comm = comm->comm_coll;

  /* Get extent of recvtype */
  MPI_Type_extent(recvtype, &extent);

  /* Lock for collective operation */
  MPID_THREAD_LOCK(comm);

  /* Send to all - receive from all */
  for ( i=0; i<size; i++ ) {
      MPI_Sendrecv ((void *)((int)sendbuf+sdispls[i]*extent), sendcnts[i], sendtype, i, MPIR_ALLTOALLV_TAG,
                    (void *)((int)recvbuf+rdispls[i]*extent), recvcnts[i], recvtype, i, MPIR_ALLTOALLV_TAG, 
		    comm, &status );
  }

  /* Unlock for collective operation */
  MPID_THREAD_UNLOCK(comm);

  return (errno);
}
